"""IO logging tests"""
import time
import json

import pytest

from pyodhean_server.app import create_app

from .common import TestingConfig
from .utils import redis_running


class TestIOLogging:

    @pytest.mark.skipif(not redis_running(), reason="This test requires Redis")
    def test_request_response_logging(
        self, json_input, tmp_path, celery_worker
    ):

        # Pass io files directory to Flask app config
        class IOLoggerTestingConfig(TestingConfig):
            IO_FILES_DIR = str(tmp_path)

        app = create_app(IOLoggerTestingConfig)
        client = app.test_client()

        # Pass io files directory to Celery app config
        celery_worker.app.conf['io_files_dir'] = str(tmp_path)

        # Post task
        response = client.post('/solver/tasks/', json=json_input)
        task_id = response.json['task_id']

        # Check request is logged
        log_files = list(tmp_path.iterdir())
        assert len(log_files) == 1
        request_file = log_files[0]
        assert request_file.name == f"{task_id}-request.json"
        assert json.load(request_file.open()) == json_input

        # Wait until task is completed
        while client.get(
                '/solver/tasks/{}/status'.format(task_id)
        ).json['status'] in ('waiting', 'ongoing'):
            time.sleep(0.1)

        # Query task result
        response = client.get('/solver/tasks/{}/result'.format(task_id))
        assert response.status_code == 200
        result = response.json
        assert result['status'] == 'ok'
        assert 'solution' in result

        # Check response is logged
        log_files = list(tmp_path.iterdir())
        assert len(log_files) == 2
        response_file = next(
            f for f in log_files if f.name == f"{task_id}-response.json"
        )
        # We can't test that response_file == response.json because the
        # response is modified by API schemas.
        assert 'solution' in json.load(response_file.open())



"""Test utils"""
import redis


def redis_running():
    """Test Redis server is available"""
    redis_client = redis.Redis(socket_connect_timeout=1)
    try:
        redis_client.ping()
    except redis.exceptions.ConnectionError:
        return False
    return True



"""PyODHeaN server application"""

from flask import Flask
from flask_smorest import Api

from pyodhean_server.solver.resources import blp as solver_blp
from pyodhean_server import logger, auth, celery
from pyodhean_server.settings import DefaultConfig
from pyodhean_server import __version__


def create_app(config_class=None):
    """Create application

    :param class config_class: Configuration class. This parameter is typically
        used for tests. In production, parameters are passed using the file in
        FLASK_SETTINGS_FILE environment variable.
    """
    app = Flask('pyodhean_server')

    # Load Flask config and override with optional settings file
    app.config.from_object(config_class or DefaultConfig)
    app.config.from_envvar('FLASK_SETTINGS_FILE', silent=True)

    # Initialize extensions
    logger.init_app(app)
    auth.init_app(app)
    celery.init_app(app)

    # Create API
    api = Api(app, spec_kwargs={"version": __version__})
    api.register_blueprint(solver_blp)

    app.logger.info("Pyodhean server started. Ready to rock.")

    return app



"""Test fixtures"""
import pyutilib.subprocess.GlobalData

import pytest

from pyodhean_server.app import create_app

from .common import TestingConfig

# https://github.com/PyUtilib/pyutilib/issues/31
# Disable signal handling
# Needed when using celery fixtures as task is launched in child thread
pyutilib.subprocess.GlobalData.DEFINE_SIGNAL_HANDLERS_DEFAULT = False


@pytest.fixture(params=[TestingConfig])
def init_app(request):
    """Initialize test application"""
    request.cls.app = create_app(request.param)
    request.cls.client = request.cls.app.test_client()


@pytest.fixture(scope='session')
def celery_config():
    return {
        'broker_url': 'redis://',
        'result_backend': 'redis://',
        # Use a dedicated queue to avoid clashing with a running worker
        'task_default_queue': 'testing',
    }


@pytest.fixture
def json_input():
    return {
        'nodes': {
            'production': [
                # P1
                {
                    'id': [0.0, 0.0],
                    'technologies': {
                        'k1': {
                            'efficiency': 0.9,
                            't_out_max': 100,
                            't_in_min': 30,
                            'production_unitary_cost': 800,
                            'energy_unitary_cost': 0.08,
                            'energy_cost_inflation_rate': 0.04,
                            'coverage_rate': 0.80,
                        },
                        'k2': {
                            'efficiency': 0.9,
                            't_out_max': 100,
                            't_in_min': 30,
                            'production_unitary_cost': 1000,
                            'energy_unitary_cost': 0.08,
                            'energy_cost_inflation_rate': 0.04,
                        },
                    },
                },
            ],
            'consumption': [
                # C1
                {
                    'id': [2.0, 5.0],
                    'kW': 80, 't_in': 60, 't_out': 80,
                },
                # C2
                {
                    'id': [30.0, 50.0],
                    'kW': 80, 't_in': 60, 't_out': 80,
                },
            ],
        },
        'links': [
            # P1 -> C1
            {'length': 10.0, 'source': [0.0, 0.0], 'target': [2.0, 5.0]},
            # C1 -> C2
            {'length': 100.0, 'source': [2.0, 5.0], 'target': [30.0, 50.0]},
        ],
        'parameters': {
            'diameter_int_max': 0.20,
            'speed_max': 2.5,
            'simultaneity_ratio': 0.70,
            'heat_loss_rate': 0.10,
        },
    }



"""Celery application

Runs PyODHeaN model solver task asynchronously
"""
import os

from celery import Celery, signals
from pyodhean_server.settings import DefaultCeleryConfig


celery = Celery(
    'PyODHeaN Solver',
    include=['pyodhean_server.solver.task'],
)
celery.config_from_object(DefaultCeleryConfig)

# https://stackoverflow.com/questions/52372117/how-to-call-a-celery-shared-task
# This seems to work around the issue of the task being owned by the Celery
# default app rather than this Celery app.
celery.set_default()

if 'CELERY_BROKER_URL' in os.environ:
    celery.conf.update(broker_url=os.environ['CELERY_BROKER_URL'])
if 'CELERY_RESULT_BACKEND' in os.environ:
    celery.conf.update(result_backend=os.environ['CELERY_RESULT_BACKEND'])
if 'CELERY_DEFAULT_QUEUE' in os.environ:
    celery.conf.update(task_default_queue=os.environ['CELERY_DEFAULT_QUEUE'])
if 'IO_FILES_DIR' in os.environ:
    celery.conf.update(io_files_dir=os.environ['IO_FILES_DIR'])


# https://stackoverflow.com/questions/9824172/find-out-whether-celery-task-exists
@signals.before_task_publish.connect
def cb_set_sent_state(sender=None, headers=None, **kwargs):
    """Set SENT custom status when the task is enqueued"""
    task = celery.tasks.get(sender)
    backend = task.backend
    backend.store_result(headers['id'], None, 'SENT')


def init_app(app):
    """Set Celery config from application"""
    celery.conf['broker_url'] = app.config['CELERY_BROKER_URL']
    celery.conf['result_backend'] = app.config['CELERY_RESULT_BACKEND']
    celery.conf['task_default_queue'] = app.config['CELERY_DEFAULT_QUEUE']



"""Authentication"""

import csv
from pathlib import Path

from werkzeug.security import generate_password_hash, check_password_hash
import click
from flask_httpauth import HTTPBasicAuth
from flask_smorest import abort

from .exceptions import PyodheanServerConfigError


auth = HTTPBasicAuth()  # pylint: disable=invalid-name


def init_app(app):
    """Initialize application authentication"""

    # Check config on startup
    if app.config["AUTH_ENABLED"]:
        if "AUTH_USERS" not in app.config:
            raise PyodheanServerConfigError("Missing AUTH_USERS parameter.")
        users_db_file = app.config["AUTH_USERS"]
        if not Path(users_db_file).exists():
            raise PyodheanServerConfigError("Missing users DB file.")

    @auth.verify_password
    def verify_password(username, password):  # pylint: disable=unused-variable
        if not app.config["AUTH_ENABLED"]:
            return True
        users = {}
        with open(app.config["AUTH_USERS"]) as users_db:
            for row in csv.reader(users_db):
                # Skip empty rows
                if row:
                    users[row[0]] = row[1]
        if username in users:
            return check_password_hash(users.get(username), password)
        return False

    @auth.error_handler
    def auth_error(status):  # pylint: disable=unused-variable
        # Call abort to trigger error handler and get consistent JSON output
        abort(status, message="Authentication error")

    app.cli.add_command(add_user)


@click.command()
@click.argument('users_db_file', type=click.File(mode='a'))
@click.argument('user')
@click.argument('password')
def add_user(users_db_file, user, password):
    """Add USER with PASSWORD to USERS_DB_FILE

    To remove a user, just edit the file and remove the line manually
    """
    users_db_file.write(f"{user},{generate_password_hash(password)}\n")



"""Authentication extension tests"""
import base64

from werkzeug.security import generate_password_hash, check_password_hash

import pytest

from pyodhean_server.app import create_app
from pyodhean_server.exceptions import PyodheanServerConfigError
from .common import TestingConfig

DUMMY_TASK_ID = '00000000-0000-0000-0000-000000000000'


class TestAuth:

    @pytest.mark.usefixtures('celery_worker')
    def test_auth_resource(self, tmp_path):
        """Test authentication feature

        Only "status" route is tested. The point is only to test the feature.
        """

        # Create users database file
        users_db_file = tmp_path / 'users.db'
        with open(users_db_file, 'w') as users_db:
            users_db.write(','.join(('test', generate_password_hash('test'))))

        class AuthTestingConfig(TestingConfig):
            AUTH_ENABLED = True
            AUTH_USERS = str(users_db_file)

        app = create_app(AuthTestingConfig)
        client = app.test_client()

        # Missing credentials
        response = client.get(
            '/solver/tasks/{}/status'.format(DUMMY_TASK_ID))
        assert response.status_code == 401
        assert response.json["message"] == "Authentication error"

        # Wrong credentials
        creds = base64.b64encode('test:wrong'.encode()).decode()
        response = client.get(
            '/solver/tasks/{}/status'.format(DUMMY_TASK_ID),
            headers={'Authorization': 'Basic ' + creds},
        )
        assert response.status_code == 401
        assert response.json["message"] == "Authentication error"

        # Right credentials
        creds = base64.b64encode('test:test'.encode()).decode()
        response = client.get(
            '/solver/tasks/{}/status'.format(DUMMY_TASK_ID),
            headers={'Authorization': 'Basic ' + creds},
        )
        assert response.status_code == 404

    def test_auth_missing_auth_users_parameter(self):
        class AuthTestingConfig(TestingConfig):
            AUTH_ENABLED = True

        with pytest.raises(
                PyodheanServerConfigError,
                match="Missing AUTH_USERS parameter."
        ):
            create_app(AuthTestingConfig)

    def test_auth_missing_auth_users_db_file(self, tmp_path):
        users_db_file = tmp_path / 'users.db'

        class AuthTestingConfig(TestingConfig):
            AUTH_ENABLED = True
            AUTH_USERS = str(users_db_file)

        with pytest.raises(
                PyodheanServerConfigError,
                match="Missing users DB file."
        ):
            create_app(AuthTestingConfig)

    @pytest.mark.usefixtures('init_app')
    def test_auth_add_user(self, tmp_path):
        users_db_file = tmp_path / 'users.db'
        result = self.app.test_cli_runner().invoke(
            args=('add-user', str(users_db_file), 'user_1', 'password_1')
        )
        assert result.exit_code == 0
        with open(users_db_file) as users_db:
            users = [line.split(',') for line in users_db.read().splitlines()]
        assert len(users) == 1
        assert check_password_hash(users[0][1], 'password_1')

        # Run twice to check line is appended
        result = self.app.test_cli_runner().invoke(
            args=('add-user', str(users_db_file), 'user_2', 'password_2')
        )
        assert result.exit_code == 0
        with open(users_db_file) as users_db:
            users = [line.split(',') for line in users_db.read().splitlines()]
        assert len(users) == 2
        assert check_password_hash(users[0][1], 'password_1')
        assert check_password_hash(users[1][1], 'password_2')



"""Resources tests"""
import time
from unittest import mock

import pytest

from kombu.exceptions import OperationalError

from pyodhean_server.solver.task import solve, CELERY_STATUSES_MAPPING

from .utils import redis_running


DUMMY_TASK_ID = '00000000-0000-0000-0000-000000000000'


class TestResources:

    @pytest.mark.usefixtures('init_app')
    def test_solve_redis_connection_error(self, json_input):
        with mock.patch.object(solve, 'delay', side_effect=OperationalError):
            response = self.client.post('/solver/tasks/', json=json_input)
        assert response.status_code == 503

    @pytest.mark.usefixtures('init_app')
    def test_solve_wrong_t_in_min_t_out_max(self, json_input):
        p1k1 = json_input['nodes']['production'][0]['technologies']['k1']
        p1k1['t_in_min'] = 90
        p1k1['t_out_max'] = 80
        response = self.client.post('/solver/tasks/', json=json_input)
        assert response.status_code == 422
        errors = response.json['errors']['json']['nodes']['production']['0'][
            'technologies']['k1']['value']['_schema']
        assert 't_in_min must be lower than t_out_max.' in errors

    @pytest.mark.usefixtures('init_app')
    def test_solve_wrong_t_in_t_out(self, json_input):
        c_1 = json_input['nodes']['consumption'][0]
        c_1['t_in'] = 90
        c_1['t_out'] = 80
        response = self.client.post('/solver/tasks/', json=json_input)
        assert response.status_code == 422
        errors = response.json['errors']['json']['nodes']['consumption']['0'][
            '_schema']
        assert 't_in must be lower than t_out.' in errors

    @pytest.mark.usefixtures('init_app')
    def test_solve_wrong_total_coverage_rate(self, json_input):
        p_1 = json_input['nodes']['production'][0]
        p_1['technologies']['k1']['coverage_rate'] = 0.9
        p_1['technologies']['k2']['coverage_rate'] = 0.5
        response = self.client.post('/solver/tasks/', json=json_input)
        assert response.status_code == 422
        errors = response.json['errors']['json']['nodes']['production']['0'][
            '_schema']
        assert (
            'Total coverage rate for a production unit must be lower than 1.'
        ) in errors

    @pytest.mark.usefixtures('init_app')
    def test_solve_link_without_node(self, json_input):
        json_input['links'].append(
            {'length': 10.0, 'source': [10.0, 10.0], 'target': [20.0, 20.0]})
        response = self.client.post('/solver/tasks/', json=json_input)
        assert response.status_code == 422
        errors = response.json['errors']['json']['_schema']
        assert 'Network contains links with no node.' in errors

    @pytest.mark.usefixtures('init_app')
    def test_status_unknown_task(self):
        with mock.patch.object(
                solve.backend, 'get_state', return_value='PENDING'
        ):
            response = self.client.get(
                '/solver/tasks/{}/status'.format(DUMMY_TASK_ID))
        assert response.status_code == 404
        assert 'Unknown task ID: {}.'.format(DUMMY_TASK_ID) in response.json[
            'message']

    @pytest.mark.parametrize(
        'status',
        (s for s in CELERY_STATUSES_MAPPING if s != 'SUCCESS')
    )
    @pytest.mark.usefixtures('init_app')
    def test_result_status_not_success(self, status):
        with mock.patch.object(
                solve.backend, 'get_state', return_value=status
        ):
            response = self.client.get(
                '/solver/tasks/{}/result'.format(DUMMY_TASK_ID))
        assert response.status_code == 404
        assert '/tasks/{}/status'.format(DUMMY_TASK_ID) in response.json[
            'message']

    @pytest.mark.skipif(not redis_running(), reason="This test requires Redis")
    @pytest.mark.usefixtures('celery_worker')
    @pytest.mark.usefixtures('init_app')
    def test_result_query_multiple_times(self, json_input):
        """Test result can be queried several times

        When using RabbitMQ as result backend (RPC backend), results are meant
        to be consumed only once: https://github.com/celery/celery/issues/5414.
        In practice, the result may be lost if queried several times.
        A reproducible case is when another task ID is queried. Then querying
        for original task ID will fail on a second attempt.

        This test serves two purposes:
        - Test the whole post request / get response procedure
        - Test the results may be queries several times using Redis
        """
        # Post task
        response = self.client.post('/solver/tasks/', json=json_input)
        task_id = response.json['task_id']
        # Wait until task is completed
        while self.client.get(
                '/solver/tasks/{}/status'.format(task_id)
        ).json['status'] in ('waiting', 'ongoing'):
            time.sleep(0.1)
        # Query result for another task ID
        response = self.client.get(
            '/solver/tasks/{}/result'.format(DUMMY_TASK_ID))
        assert response.status_code == 404
        # Query task result
        response = self.client.get('/solver/tasks/{}/result'.format(task_id))
        assert response.status_code == 200
        result = response.json
        assert result['status'] == 'ok'
        assert 'solution' in result
        # Query task result again
        response = self.client.get('/solver/tasks/{}/result'.format(task_id))
        assert response.status_code == 200
        assert response.json == result



"""PyODHeaN solver task"""
import pathlib
import json

from celery import shared_task
from celery.utils.log import get_task_logger

from pyodhean.interface import JSONInterface


logger = get_task_logger(__name__)


# See "Options available via the AMPL Interface" in IPOPT documentation.
# https://www.coin-or.org/Ipopt/documentation/node63.html
OPTIONS = {
    'tol': 1e-3,
}


# Celery app sets status to SENT when a task is published.
# A PENDING status means the task is unknown (wrong ID or deleted task)
# We don't need PENDING in the mapping because the view returns 404 on PENDING
CELERY_STATUSES_MAPPING = {
    'SENT': 'waiting',
    'RETRY': 'waiting',
    'STARTED': 'ongoing',
    'FAILURE': 'server_error',
    'REVOKED': 'server_error',
    'SUCCESS': 'success',
}


SOLVER_STATUSES_MAPPING = {
    'ok': 'ok',
    'warning': 'no_solution_found',
    'error': 'solver_error',
    'aborted': 'solver_error',
    'unknown': 'solver_error'
}


@shared_task(bind=True)
def solve(self, json_input):
    """Solve PyODHeaN model"""
    solver = JSONInterface(OPTIONS)
    logger.info('Starting solver')
    json_output = solver.solve(json_input, tee=False, keepfiles=False)
    status = json_output['status']
    if status in ('error', 'aborted', 'unknown'):
        logger.error(
            'Solver failed with status %(status)s', {'status': status})
    else:
        logger.info(
            'Solver succeeded with status %(status)s', {'status': status})
    json_output['status'] = SOLVER_STATUSES_MAPPING[status]
    # Log response to file
    io_files_dir = self.app.conf.get('io_files_dir')
    if io_files_dir is not None:
        with (
            pathlib.Path(io_files_dir) / f"{self.request.id}-response.json"
        ).open("w") as json_file:
            json_file.write(json.dumps(json_output, indent=2))
    return json_output



"""Application logger"""

import time
from pathlib import Path
import logging
from logging.handlers import WatchedFileHandler
from flask import request
from flask.logging import default_handler


class RequestFormatter(logging.Formatter):
    """Custom log Formatter injecting request information

    http://flask.pocoo.org/docs/1.0/logging/#injecting-request-information
    """
    def format(self, record):
        if request:
            record.url = request.url
            record.method = request.method
            record.remote_addr = request.remote_addr
        else:
            record.url = ''
            record.method = ''
            record.remote_addr = ''
        return super().format(record)


def init_app(app):
    """Initialize application logger"""

    # Use default config in DEBUG mode, configure logger otherwise
    if not app.debug and app.config['LOGGER_ENABLED']:

        log_level = app.config['LOGGER_LEVEL']
        log_dir = Path(app.config['LOGGER_DIR'])
        log_format = app.config['LOGGER_FORMAT']

        # Remove Flask default handler
        app.logger.removeHandler(default_handler)

        # Use WatchedFileHandler to reopen if logrotate rotates the log file
        file_handler = WatchedFileHandler(log_dir / 'pyodhean.log')

        # Create record formatter
        formatter = RequestFormatter(log_format)
        formatter.converter = time.gmtime
        file_handler.setFormatter(formatter)

        # Add our custom handler
        app.logger.addHandler(file_handler)

        # Set logging level
        app.logger.setLevel(log_level)



"""Solver API Blueprint"""
import pathlib
import json

from flask import current_app
from kombu.exceptions import OperationalError

from flask_smorest import Blueprint, abort

from pyodhean_server.auth import auth
from .task import solve, CELERY_STATUSES_MAPPING
from .schemas import (
    SolverInputSchema, SolverOutputSchema, TaskIdSchema, StatusSchema)


blp = Blueprint(
    'Solver', __name__, url_prefix='/solver', description="Solver resources")


@blp.route('/tasks/', methods=['POST'])
@auth.login_required
@blp.arguments(SolverInputSchema)
@blp.response(TaskIdSchema)
def enqueue(json_input):
    """Enqueue solver task"""
    try:
        task = solve.delay(json_input)
    except OperationalError:
        current_app.logger.error("Can't queue task. Check Redis is running.")
        abort(503)
    current_app.logger.info("Task %(task_id)s enqueued.", {'task_id': task.id})
    current_app.logger.debug(json.dumps(json_input))
    # Log request to file
    io_files_dir = current_app.config.get('IO_FILES_DIR')
    if io_files_dir is not None:
        with (
            pathlib.Path(io_files_dir) / f"{task.id}-request.json"
        ).open("w") as json_file:
            json_file.write(json.dumps(json_input, indent=2))
    return task


@blp.route('/tasks/<uuid:task_id>/status')
@auth.login_required
@blp.response(StatusSchema)
def status(task_id):
    """Get solver task status"""
    task_state = solve.backend.get_state(str(task_id))
    if task_state == 'PENDING':
        abort(
            404,
            message=(
                'Unknown task ID: {}. '
                'Either the task ID is wrong or the task was deleted.'
                ''.format(task_id)
            )
        )
    return {'status': CELERY_STATUSES_MAPPING[task_state]}


@blp.route('/tasks/<uuid:task_id>/result')
@auth.login_required
@blp.response(SolverOutputSchema)
def result(task_id):
    """Get solver task result"""
    task_state = solve.backend.get_state(str(task_id))
    if task_state != 'SUCCESS':
        abort(
            404,
            message=(
                'Task result unavailable because status is not "success". '
                'Please check status first: /tasks/{}/status'.format(task_id)
            )
        )
    return solve.AsyncResult(str(task_id)).result



"""Default application settings"""


class DefaultConfig:
    """Default configuration"""

    API_TITLE = 'PyODHeaN server'

    # Authentication
    AUTH_ENABLED = False

    # Logging
    LOGGER_ENABLED = False
    LOGGER_LEVEL = 'WARNING'
    LOGGER_FORMAT = (
        '%(asctime)s | %(levelname)-8s | '
        '%(remote_addr)-15s | %(method)-7s | %(url)s | %(message)s'
    )

    # Celery settings
    CELERY_BROKER_URL = 'redis://'
    CELERY_RESULT_BACKEND = 'redis://'
    CELERY_DEFAULT_QUEUE = 'pyodhean'

    # OpenAPI settings
    # These settings are used by the automatic swagger API documentation.
    # They may be overidden to disable it or configure it differently.
    # In production, it is a good idea to pin the ReDoc version number rather
    # than use "next".
    # See flask-smorest documentation:
    # https://flask-smorest.readthedocs.io/en/latest/openapi.html
    OPENAPI_VERSION = '3.0.2'
    OPENAPI_URL_PREFIX = '/'
    OPENAPI_REDOC_PATH = '/'
    OPENAPI_REDOC_URL = (
        "https://cdn.jsdelivr.net/npm/redoc@next/bundles/redoc.standalone.js"
    )


class DefaultCeleryConfig:
    """Default Celery configuration"""
    broker_url = 'redis://'
    result_backend = 'redis://'
    task_default_queue = 'pyodhean'
    task_track_started = True



#!/usr/bin/env python3
"""PyODHeaN solver server

Optimization of District Heating Networks

This package provides the solver server.
"""

from setuptools import setup, find_packages

# Get the long description from the README file
with open("README.rst", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="pyodhean-server",
    version="1.0.0b1",
    description="Optimization of District Heating Networks",
    long_description=long_description,
    url="https://github.com/sigopti/pyodhean-server",
    author="Nobatek/INEF4",
    author_email="jlafrechoux@nobatek.com",
    license="GPLv3+",
    keywords=[
        "District",
        "Heating",
        "Network",
        "Optimization",
    ],
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering",
        (
            "License :: OSI Approved :: "
            "GNU Affero General Public License v3 or later (AGPLv3+)"
        ),
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    python_requires=">=3.6",
    install_requires=[
        "pyodhean>=1.0.0b1",
        "celery[redis]>=4.2.0,<5.0",
        "werkzeug>=0.15",
        "flask>=1.0.0",
        "python-dotenv>=0.9.0",
        "flask-smorest>=0.23.0,<0.24",
        "marshmallow>=3.0.0",
        "flask_httpauth>=3.0.0"
    ],
    packages=find_packages(exclude=["tests*"]),
)



"""Solver module schemas"""
# pylint: disable=missing-docstring

import marshmallow as ma

from .task import CELERY_STATUSES_MAPPING, SOLVER_STATUSES_MAPPING


class Schema(ma.Schema):
    class Meta:
        ordered = True


class NodeSchema(Schema):
    id = ma.fields.List(
        ma.fields.Float,
        validate=ma.validate.Length(min=2, max=2),
        required=True,
    )


class LinkSchema(Schema):
    source = ma.fields.List(
        ma.fields.Float,
        validate=ma.validate.Length(min=2, max=2),
        required=True,
    )
    target = ma.fields.List(
        ma.fields.Float,
        validate=ma.validate.Length(min=2, max=2),
        required=True,
    )


class InputProductionTechnologySchema(Schema):

    @ma.validates_schema
    def validate_t_in_min_t_out_max(self, data, **kwargs):
        # pylint: disable=unused-argument,no-self-use
        if data['t_in_min'] > data['t_out_max']:
            raise ma.ValidationError("t_in_min must be lower than t_out_max.")

    efficiency = ma.fields.Float(
        required=True,
        validate=ma.validate.Range(min=0),
    )
    t_out_max = ma.fields.Float(
        required=True,
    )
    t_in_min = ma.fields.Float(
        required=True,
    )
    production_unitary_cost = ma.fields.Float(
        required=True,
        validate=ma.validate.Range(min=0),
    )
    energy_unitary_cost = ma.fields.Float(
        required=True,
        validate=ma.validate.Range(min=0),
    )
    energy_cost_inflation_rate = ma.fields.Float(
        required=True,
    )
    coverage_rate = ma.fields.Float(
        validate=ma.validate.Range(min=0, max=1),
    )


class InputProductionNodeSchema(NodeSchema):

    @ma.validates_schema
    def validate_coverage_rates(self, data, **kwargs):
        # pylint: disable=unused-argument,no-self-use
        if sum(
                t.get('coverage_rate', 0)
                for t in data['technologies'].values()
        ) > 1:
            raise ma.ValidationError(
                "Total coverage rate for a production unit "
                "must be lower than 1."
            )

    technologies = ma.fields.Dict(
        ma.fields.String(),
        ma.fields.Nested(InputProductionTechnologySchema),
        required=True,
    )


class InputConsumptionNodeSchema(NodeSchema):

    @ma.validates_schema
    def validate_t_in_t_out(self, data, **kwargs):
        # pylint: disable=unused-argument,no-self-use
        if data['t_in'] > data['t_out']:
            raise ma.ValidationError("t_in must be lower than t_out.")

    kW = ma.fields.Float(
        required=True
    )
    t_in = ma.fields.Float(
        required=True
    )
    t_out = ma.fields.Float(
        required=True
    )


class InputNodesSchema(Schema):
    production = ma.fields.List(
        ma.fields.Nested(InputProductionNodeSchema),
        required=True
    )
    consumption = ma.fields.List(
        ma.fields.Nested(InputConsumptionNodeSchema),
        required=True
    )


class InputLinkSchema(LinkSchema):
    length = ma.fields.Float(
        required=True
    )


class InputParametersSchema(Schema):

    speed_max = ma.fields.Float(
        validate=ma.validate.Range(min=0),
    )
    diameter_int_max = ma.fields.Float(
        validate=ma.validate.Range(min=0),
    )
    operation_time = ma.fields.Float(
        validate=ma.validate.Range(min=0, max=365*24),
    )
    depreciation_period = ma.fields.Float(
        validate=ma.validate.Range(min=0),
    )
    discout_rate = ma.fields.Float(
        validate=ma.validate.Range(min=0),
    )
    trench_unit_cost = ma.fields.Float(
        validate=ma.validate.Range(min=0),
    )
    pipe_diameter_unit_cost_slope = ma.fields.Float()
    pipe_diameter_unit_cost_y_intercept = ma.fields.Float()
    exchanger_power_cost_slope = ma.fields.Float()
    exchanger_power_cost_y_intercept = ma.fields.Float()
    pump_energy_ratio_cost = ma.fields.Float(
        validate=ma.validate.Range(min=0, max=1),
    )
    simultaneity_ratio = ma.fields.Float(
        validate=ma.validate.Range(min=0, max=1),
    )
    heat_loss_rate = ma.fields.Float(
        validate=ma.validate.Range(min=0, max=1),
    )


class SolverInputSchema(Schema):

    @ma.validates_schema
    def validate_links_coordinates(self, data, **kwargs):
        # pylint: disable=unused-argument,no-self-use
        """Check all links link to nodes"""
        node_coords = (
            set(tuple(p['id']) for p in data['nodes']['production']) |
            set(tuple(c['id']) for c in data['nodes']['consumption'])
        )
        link_coords = (
            set(tuple(link['source']) for link in data['links']) |
            set(tuple(link['target']) for link in data['links'])
        )
        if link_coords - node_coords:
            raise ma.ValidationError("Network contains links with no node.")

    nodes = ma.fields.Nested(
        InputNodesSchema,
        required=True
    )
    links = ma.fields.List(
        ma.fields.Nested(InputLinkSchema),
        required=True
    )
    parameters = ma.fields.Nested(InputParametersSchema)


class OutputProductionTechnologySchema(Schema):
    flow_rate = ma.fields.Float(
        required=True
    )
    power = ma.fields.Float(
        required=True
    )
    t_supply = ma.fields.Float(
        required=True
    )
    t_return = ma.fields.Float(
        required=True
    )


class OutputProductionNodeSchema(NodeSchema):
    technologies = ma.fields.Dict(
        ma.fields.String(),
        ma.fields.Nested(OutputProductionTechnologySchema),
        required=True,
    )
    t_supply = ma.fields.Float(
        required=True
    )
    t_return = ma.fields.Float(
        required=True
    )
    pump_pressure = ma.fields.Float(
        required=True
    )
    flow_rate = ma.fields.Float(
        required=True
    )


class OutputConsumptionNodeSchema(NodeSchema):
    flow_rate_before_exchanger = ma.fields.Float(
        required=True
    )
    flow_rate_after_exchanger = ma.fields.Float(
        required=True
    )
    flow_rate_in_exchanger = ma.fields.Float(
        required=True
    )
    exchanger_power = ma.fields.Float(
        required=True
    )
    exchanger_surface = ma.fields.Float(
        required=True
    )
    exchanger_t_in = ma.fields.Float(
        required=True
    )
    exchanger_t_out = ma.fields.Float(
        required=True
    )
    exchanger_t_supply = ma.fields.Float(
        required=True
    )
    exchanger_t_return = ma.fields.Float(
        required=True
    )
    exchanger_DTLM = ma.fields.Float(
        required=True
    )
    exchanger_delta_t_cold = ma.fields.Float(
        required=True
    )
    exchanger_delta_t_hot = ma.fields.Float(
        required=True
    )


class OutputNodesSchema(Schema):
    production = ma.fields.List(
        ma.fields.Nested(OutputProductionNodeSchema),
        required=True
    )
    consumption = ma.fields.List(
        ma.fields.Nested(OutputConsumptionNodeSchema),
        required=True
    )


class OutputLinkSchema(LinkSchema):
    diameter_int = ma.fields.Float(
        required=True
    )
    diameter_out = ma.fields.Float(
        required=True
    )
    flow_rate = ma.fields.Float(
        required=True
    )
    speed = ma.fields.Float(
        required=True
    )
    t_return_in = ma.fields.Float(
        required=True
    )
    t_return_out = ma.fields.Float(
        required=True
    )
    t_supply_in = ma.fields.Float(
        required=True
    )
    t_supply_out = ma.fields.Float(
        required=True
    )
    power = ma.fields.Float(
        required=True
    )
    yearly_energy = ma.fields.Float(
        required=True
    )


class OutputGlobalIndicatorsSchema(Schema):
    production_intallation_cost = ma.fields.Float(
        required=True
    )
    exchangers_installation_cost = ma.fields.Float(
        required=True
    )
    network_cost = ma.fields.Float(
        required=True
    )
    trenches_cost = ma.fields.Float(
        required=True
    )
    pipes_cost = ma.fields.Float(
        required=True
    )
    network_length = ma.fields.Float(
        required=True
    )
    heat_production_cost = ma.fields.Float(
        required=True
    )
    pumps_operation_cost = ma.fields.Float(
        required=True
    )
    total_capex = ma.fields.Float(
        required=True
    )
    total_opex = ma.fields.Float(
        required=True
    )
    total_cost = ma.fields.Float(
        required=True
    )
    yearly_production = ma.fields.Float(
        required=True
    )
    total_production = ma.fields.Float(
        required=True
    )


class OutputSolutionSchema(Schema):
    global_indicators = ma.fields.Nested(
        OutputGlobalIndicatorsSchema,
        required=True
    )
    nodes = ma.fields.Nested(
        OutputNodesSchema,
        required=True
    )
    links = ma.fields.List(
        ma.fields.Nested(OutputLinkSchema),
        required=True
    )


class SolverOutputSchema(Schema):
    solution = ma.fields.Nested(OutputSolutionSchema)
    status = ma.fields.String(
        validate=ma.validate.OneOf(SOLVER_STATUSES_MAPPING.values()),
        required=True
    )


class StatusSchema(Schema):
    status = ma.fields.String(
        validate=ma.validate.OneOf(CELERY_STATUSES_MAPPING.values()),
        required=True
    )


class TaskIdSchema(Schema):
    task_id = ma.fields.UUID(
        attribute='id',
        required=True
    )
